from __future__ import annotations

import os
import shlex
import sys
from pathlib import Path
from subprocess import run
from typing import Any, Literal
from webbrowser import open as open_url

import click
from django.core.management import BaseCommand, get_commands, load_command_class
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.text import Text
from textual import events, on
from textual.app import App, AutopilotCallbackType, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.css.query import NoMatches
from textual.screen import Screen
from textual.widgets import (
    Button,
    Footer,
    Label,
    Static,
    Tree,
)
from textual.widgets.tree import TreeNode
from trogon.introspect import ArgumentSchema, CommandSchema, MultiValueParamData, OptionSchema
from trogon.run_command import UserCommandData
from trogon.widgets.about import TextDialog
from trogon.widgets.command_info import CommandInfo
from trogon.widgets.command_tree import CommandTree
from trogon.widgets.form import CommandForm
from trogon.widgets.multiple_choice import NonFocusableVerticalScroll

from django_tui.management.commands.ish import InteractiveShellScreen


def introspect_django_commands() -> dict[str, CommandSchema]:
    groups = {}
    for name, app_name in get_commands().items():
        try:
            kls = load_command_class(app_name, name)
        except AttributeError:
            # Skip invalid commands
            continue
        if app_name == "django.core":
            group_name = "django"
        else:
            group_name = app_name.rpartition(".")[-1]

        parser = kls.create_parser(f"django {name}", name)
        options = []
        args = []
        root = []
        for action in parser._actions:
            if action.nargs == "?":
                nargs = 1
            elif action.nargs in ("*", "+"):
                nargs = -1
            elif not action.nargs:
                nargs = 1
            else:
                nargs = action.nargs

            if hasattr(action, "type"):
                if action.type is bool:
                    type_ = click.BOOL
                elif action.type is int:
                    type_ = click.INT
                elif action.type is str:
                    type_ = click.STRING
                else:
                    type_ = click.STRING if action.nargs != 0 else click.BOOL
            else:
                type_ = click.STRING if action.nargs != 0 else click.BOOL

            default = action.default
            if default is None:
                default = MultiValueParamData([])
            elif type_ is click.BOOL:
                default = MultiValueParamData([])
            else:
                default = MultiValueParamData(values=[(default,)])

            if not action.option_strings:
                args.append(
                    ArgumentSchema(
                        name=action.metavar or action.dest,
                        type=type_,
                        required=action.required if action.nargs != "*" else False,
                        default=default,
                        choices=action.choices,
                        multiple=action.nargs in ("+", "*"),
                        nargs=nargs,
                    )
                )
                continue
            option_name = action.option_strings[0]

            schema = OptionSchema(
                name=option_name,
                type=type_,
                help=action.help,
                default=default,
                required=action.required,
                multiple=action.nargs in ("+", "*"),
                choices=action.choices,
                is_flag=action.nargs == 0,
                is_boolean_flag=action.nargs == 0,
                nargs=nargs,
            )
            if option_name in (
                "-h",
                "--version",
                "-v",
                "--settings",
                "--pythonpath",
                "--traceback",
                "--no-color",
                "--force-color",
                "--skip-checks",
            ):
                root.append(schema)
            else:
                options.append(schema)

        if group_name not in groups:
            groups[group_name] = CommandSchema(name=group_name, function=None, is_group=True, options=root)

        command = CommandSchema(
            name=name,
            function=None,
            is_group=False,
            docstring=None,
            options=options,
            arguments=args,
            parent=groups[group_name],
        )

        groups[group_name].subcommands[name] = command

    return groups


class AboutDialog(TextDialog):
    DEFAULT_CSS = """
    TextDialog > Vertical {
        border: thick $primary 50%;
    }
    """

    def __init__(self) -> None:
        title = f"About django-tui"
        message = Text.from_markup(
            "Built with [@click=app.visit('https://github.com/textualize/textual')]Textual[/] & [@click=app.visit('https://github.com/textualize/trogon')]Trogon[/] "
            "by [@click=app.visit('https://pecar.me')]Anže Pečar[/].\n\n"
            "Interactive Shell contributed by [@click=app.visit('https://github.com/shtayeb')]Shahryar Tayeb[/].\n\n"
            "[@click=app.visit('https://github.com/anze3db/django-tui')]"
            "https://github.com/anze3db/django-tui[/]",
        )
        super().__init__(title, message)


# 2 For the command screen
class DjangoCommandBuilder(Screen):
    COMPONENT_CLASSES = {"version-string", "prompt", "command-name-syntax"}

    BINDINGS = [
        Binding(key="ctrl+r", action="close_and_run", description="Close & Run"),
        Binding(key="ctrl+z", action="copy_command", description="Copy Command to Clipboard"),
        Binding(key="ctrl+t", action="focus_command_tree", description="Focus Command Tree"),
        # Binding(key="ctrl+o", action="show_command_info", description="Command Info"),
        Binding(key="ctrl+s", action="focus('search')", description="Search"),
        Binding(key="ctrl+j", action="select_mode('shell')", description="Shell"),
        ("escape", "app.back", "Back"),
        Binding(key="f1", action="about", description="About"),
    ]

    def __init__(
        self,
        click_app_name: str,
        command_name: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        super().__init__(name, id, classes)
        self.command_data = None
        self.is_grouped_cli = True

        self.command_schemas = introspect_django_commands()
        self.click_app_name = click_app_name
        self.command_name = command_name

        try:
            self.version = metadata.version(self.click_app_name)
        except Exception:
            self.version = None

        self.highlighter = ReprHighlighter()

    def compose(self) -> ComposeResult:
        tree = CommandTree("Commands", self.command_schemas, self.command_name)

        title_parts = [Text(self.click_app_name, style="b")]
        if self.version:
            version_style = self.get_component_rich_style("version-string")
            title_parts.extend(["\n", (f"v{self.version}", version_style)])

        title = Text.assemble(*title_parts)

        sidebar = Vertical(
            Label(title, id="home-commands-label"),
            tree,
            id="home-sidebar",
        )
        if self.is_grouped_cli:
            # If the root of the click app is a Group instance, then
            #  we display the command tree to users and focus it.
            tree.focus()
        else:
            # If the click app is structured using a single command,
            #  there's no need for us to display the command tree.
            sidebar.display = False

        yield sidebar

        with Vertical(id="home-body"):
            with Horizontal(id="home-command-description-container") as vs:
                vs.can_focus = False
                yield Static(self.click_app_name or "", id="home-command-description")

            scrollable_body = VerticalScroll(
                Static(""),
                id="home-body-scroll",
            )
            scrollable_body.can_focus = False
            yield scrollable_body
            yield Horizontal(
                NonFocusableVerticalScroll(
                    Static("", id="home-exec-preview-static"),
                    id="home-exec-preview-container",
                ),
                # Vertical(
                #     Button.success("Close & Run", id="home-exec-button"),
                #     id="home-exec-preview-buttons",
                # ),
                id="home-exec-preview",
            )

        yield Footer()

    def action_close_and_run(self) -> None:
        self.app.execute_on_exit = True
        self.app.exit()

    def action_copy_command(self) -> None:
        if self.command_data is None:
            return

        if sys.platform == "win32":
            copy_command = ["clip"]
        elif sys.platform == "darwin":
            copy_command = ["pbcopy"]
        else:
            copy_command = ["xclip", "-selection", "clipboard"]

        try:
            command = self.click_app_name + " " + " ".join(shlex.quote(str(x)) for x in self.command_data.to_cli_args())
            run(
                copy_command,
                input=command,
                text=True,
                check=False,
            )
            self.notify(f"`{command}` copied to clipboard.")
        except FileNotFoundError:
            self.notify(f"Could not copy to clipboard. `{copy_command[0]}` not found.", severity="error")

    def action_about(self) -> None:
        self.app.push_screen(AboutDialog())

    async def on_mount(self, event: events.Mount) -> None:
        await self._refresh_command_form()

    async def _refresh_command_form(self, node: TreeNode[CommandSchema] | None = None):
        if node is None:
            try:
                command_tree = self.query_one(CommandTree)
                node = command_tree.cursor_node
            except NoMatches:
                return

        self.selected_command_schema = node.data
        self._update_command_description(node)
        self._update_execution_string_preview(self.selected_command_schema, self.command_data)
        await self._update_form_body(node)

    @on(Tree.NodeHighlighted)
    async def selected_command_changed(self, event: Tree.NodeHighlighted[CommandSchema]) -> None:
        """When we highlight a node in the CommandTree, the main body of the home page updates
        to display a form specific to the highlighted command."""
        await self._refresh_command_form(event.node)

    @on(CommandForm.Changed)
    def update_command_data(self, event: CommandForm.Changed) -> None:
        self.command_data = event.command_data
        self._update_execution_string_preview(self.selected_command_schema, self.command_data)

    def _update_command_description(self, node: TreeNode[CommandSchema]) -> None:
        """Update the description of the command at the bottom of the sidebar
        based on the currently selected node in the command tree."""
        description_box = self.query_one("#home-command-description", Static)
        description_text = node.data.docstring or ""
        description_text = description_text.lstrip()
        description_text = f"[b]{node.label if self.is_grouped_cli else self.click_app_name}[/]\n{description_text}"
        description_box.update(description_text)

    def _update_execution_string_preview(self, command_schema: CommandSchema, command_data: UserCommandData) -> None:
        """Update the preview box showing the command string to be executed"""
        if self.command_data is not None:
            command_name_syntax_style = self.get_component_rich_style("command-name-syntax")
            prefix = Text(f"{self.click_app_name} ", command_name_syntax_style)
            new_value = command_data.to_cli_string(include_root_command=False)
            highlighted_new_value = Text.assemble(prefix, self.highlighter(new_value))
            prompt_style = self.get_component_rich_style("prompt")
            preview_string = Text.assemble(("$ ", prompt_style), highlighted_new_value)
            self.query_one("#home-exec-preview-static", Static).update(preview_string)

    async def _update_form_body(self, node: TreeNode[CommandSchema]) -> None:
        # self.query_one(Pretty).update(node.data)
        parent = self.query_one("#home-body-scroll", VerticalScroll)
        for child in parent.children:
            await child.remove()

        # Process the metadata for this command and mount corresponding widgets
        command_schema = node.data
        command_form = CommandForm(command_schema=command_schema, command_schemas=self.command_schemas)
        await parent.mount(command_form)
        if not self.is_grouped_cli:
            command_form.focus()


class DjangoTui(App):
    CSS_PATH = Path(__file__).parent / "trogon.scss"

    def __init__(
        self,
        *,
        open_shell: bool = False,
    ) -> None:
        super().__init__()
        self.post_run_command: list[str] = []
        self.is_grouped_cli = True
        self.execute_on_exit = False
        self.app_name = "python manage.py"
        self.command_name = "django-tui"
        self.open_shell = open_shell

    def on_mount(self):
        if self.open_shell:
            self.push_screen(InteractiveShellScreen("Interactive Shell"))
        else:
            self.push_screen(DjangoCommandBuilder(self.app_name, self.command_name))
            # self.push_screen(HomeScreen(self.app_name))

    @on(Button.Pressed, "#home-exec-button")
    def on_button_pressed(self):
        self.execute_on_exit = True
        self.exit()

    def run(
        self,
        *,
        headless: bool = False,
        size: tuple[int, int] | None = None,
        auto_pilot: AutopilotCallbackType | None = None,
    ) -> None:
        try:
            super().run(headless=headless, size=size, auto_pilot=auto_pilot)
        finally:
            if self.post_run_command:
                console = Console()
                if self.post_run_command and self.execute_on_exit:
                    console.print(
                        f"Running [b cyan]{self.app_name} {' '.join(shlex.quote(s) for s in self.post_run_command)}[/]"
                    )

                    split_app_name = shlex.split(self.app_name)
                    program_name = shlex.split(self.app_name)[0]
                    arguments = [*split_app_name, *self.post_run_command]
                    os.execvp(program_name, arguments)

    @on(CommandForm.Changed)
    def update_command_to_run(self, event: CommandForm.Changed):
        include_root_command = not self.is_grouped_cli
        self.post_run_command = event.command_data.to_cli_args(include_root_command)

    def action_focus_command_tree(self) -> None:
        try:
            command_tree = self.query_one(CommandTree)
        except NoMatches:
            return

        command_tree.focus()

    def action_show_command_info(self) -> None:
        command_builder = self.query_one(DjangoCommandBuilder)
        self.push_screen(CommandInfo(command_builder.selected_command_schema))

    def action_visit(self, url: str) -> None:
        """Visit the given URL, via the operating system.

        Args:
            url: The URL to visit.
        """
        open_url(url)

    def action_select_mode(self, mode_id: Literal["commands", "shell"]) -> None:
        if mode_id == "commands":
            self.app.push_screen(DjangoCommandBuilder("pyhton manage.py", "Test command name"))

        elif mode_id == "shell":
            self.app.push_screen(InteractiveShellScreen("Interactive Shell"))


class Command(BaseCommand):
    help = """Run and inspect Django commands in a text-based user interface (TUI)."""

    def add_arguments(self, parser):
        parser.add_argument("--shell", action="store_true", help="Open django shell")

    def handle(self, *args: Any, shell=False, **options: Any) -> None:
        app = DjangoTui(open_shell=shell)
        app.run()



from __future__ import annotations

import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))
DEBUG = False
TEMPLATE_DEBUG = DEBUG

SECRET_KEY = "THISuISdNOT9A$SECRET9x&ji!vceayg+wwt472!bgs$0!i3k4"

DATABASES = {
    "default": {
        "ENGINE": "tests.db_backend",
        "NAME": "django_mysql",
        "USER": os.environ.get("DB_USER", ""),
        "PASSWORD": os.environ.get("DB_PASSWORD", ""),
        "HOST": os.environ.get("DB_HOST", ""),
        "PORT": os.environ.get("DB_PORT", ""),
        "OPTIONS": {"charset": "utf8mb4"},
        "TEST": {"COLLATION": "utf8mb4_general_ci", "CHARSET": "utf8mb4"},
    },
    "other": {
        "ENGINE": "tests.db_backend",
        "NAME": "django_mysql2",
        "USER": os.environ.get("DB_USER", ""),
        "PASSWORD": os.environ.get("DB_PASSWORD", ""),
        "HOST": os.environ.get("DB_HOST", ""),
        "PORT": os.environ.get("DB_PORT", ""),
        "OPTIONS": {"charset": "utf8mb4"},
        "TEST": {"COLLATION": "utf8mb4_general_ci", "CHARSET": "utf8mb4"},
    },
    "other2": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    },
}

DATABASE_ROUTERS = ["tests.testapp.routers.NothingOnSQLiteRouter"]

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

ALLOWED_HOSTS: list[str] = []

INSTALLED_APPS = [
    "tests.testapp",
    "django_mysql",
    "django.contrib.contenttypes",
]

ROOT_URLCONF = "tests.urls"
TIME_ZONE = "UTC"
USE_I18N = True
USE_TZ = True

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "OPTIONS": {"context_processors": ["django.contrib.auth.context_processors.auth"]},
    }
]

DJANGO_MYSQL_REWRITE_QUERIES = True



from __future__ import annotations

import importlib
import os
import platform
import sys
import traceback
import warnings
from functools import lru_cache
from io import StringIO
from subprocess import run
from typing import List, Tuple

import django
from django.apps import apps
from rich.syntax import Syntax
from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import HorizontalScroll, Vertical, VerticalScroll
from textual.screen import ModalScreen, Screen
from textual.widgets import (
    Footer,
    Label,
    MarkdownViewer,
    TextArea,
)
from textual.widgets.text_area import Location, Selection

DEFAULT_IMPORT = {
    "rich": ["print_json", "print"],
    "django.db.models": [
        "Avg",
        "Case",
        "Count",
        "F",
        "Max",
        "Min",
        "Prefetch",
        "Q",
        "Sum",
        "When",
    ],
    "django.conf": [
        "settings",
    ],
    "django.core.cache": [
        "cache",
    ],
    "django.contrib.auth": [
        "get_user_model",
    ],
    "django.utils": [
        "timezone",
    ],
    "django.urls": ["reverse"],
}


@lru_cache
def get_modules():
    """
    Return list of modules and symbols to import
    """
    mods = {}
    for module_name, symbols in DEFAULT_IMPORT.items():
        try:
            module = importlib.import_module(module_name)
        except ImportError as e:
            warnings.warn("django_admin_shell - autoimport warning :: {msg}".format(msg=str(e)), ImportWarning)
            continue

        mods[module_name] = []
        for symbol_name in symbols:
            if hasattr(module, symbol_name):
                mods[module_name].append(symbol_name)
            else:
                warnings.warn(
                    "django_admin_shell - autoimport warning :: "
                    "AttributeError module '{mod}' has no attribute '{attr}'".format(mod=module_name, attr=symbol_name),
                    ImportWarning,
                )

    for model_class in apps.get_models():
        _mod = model_class.__module__
        classes = mods.get(_mod, [])
        classes.append(model_class.__name__)
        mods[_mod] = classes

    return mods


@lru_cache
def get_scope():
    """
    Return map with symbols to module/object
    Like:
    "reverse" -> "django.urls.reverse"
    """
    scope = {}
    for module_name, symbols in get_modules().items():
        module = importlib.import_module(module_name)
        for symbol_name in symbols:
            scope[symbol_name] = getattr(module, symbol_name)
    return scope


@lru_cache
def import_str():
    buf = []
    for module, symbols in get_modules().items():
        if symbols:
            buf.append(f"from {module} import {', '.join(symbols)}")
    return "\n".join(buf)


def run_code(code):
    """
    Execute code and return result with status = success|error
    Function manipulate stdout to grab output from exec
    """
    status = "success"
    out = ""
    tmp_stdout = sys.stdout
    buf = StringIO()

    try:
        sys.stdout = buf
        exec(code, None, get_scope())
    except Exception:
        out = traceback.format_exc()
        status = "error"
    else:
        out = buf.getvalue()
    finally:
        sys.stdout = tmp_stdout

    result = {
        "code": code,
        "out": out,
        "status": status,
    }
    return result


class ExtendedTextArea(TextArea):
    """A subclass of TextArea with parenthesis-closing functionality."""

    def _on_key(self, event: events.Key) -> None:
        auto_close_chars = {
            "(": ")",
            "{": "}",
            "[": "]",
            "'": "'",
            '"': '"',
        }
        closing_char = auto_close_chars.get(event.character)
        if closing_char:
            self.insert(f"{event.character}{closing_char}")
            self.move_cursor_relative(columns=-1)
            event.prevent_default()


class TextEditorBindingsInfo(ModalScreen[None]):
    BINDINGS = [
        Binding("escape", "dismiss(None)", "", show=False),
    ]

    DEFAULT_CSS = """
    TextEditorBindingsInfo {
        align: center middle;
    }
"""

    key_bindings = """
Text Editor Key Bindings List
| Key(s)           | Description                                          |
|------------------|------------------------------------------------------|
| escape           | Focus on the next item.                              |
| up               | Move the cursor up.                                  |
| down             | Move the cursor down.                                |
| left             | Move the cursor left.                                |
| ctrl+left        | Move the cursor to the start of the word.            |
| ctrl+shift+left  | Move the cursor to the start of the word and select. |
| right            | Move the cursor right.                               |
| ctrl+right       | Move the cursor to the end of the word.              |
| ctrl+shift+right | Move the cursor to the end of the word and select.   |
| home,ctrl+a      | Move the cursor to the start of the line.            |
| end,ctrl+e       | Move the cursor to the end of the line.              |
| shift+home       | Move the cursor to the start of the line and select. |
| shift+end        | Move the cursor to the end of the line and select.   |
| pageup           | Move the cursor one page up.                         |
| pagedown         | Move the cursor one page down.                       |
| shift+up         | Select while moving the cursor up.                   |
| shift+down       | Select while moving the cursor down.                 |
| shift+left       | Select while moving the cursor left.                 |
| shift+right      | Select while moving the cursor right.                |
| backspace        | Delete character to the left of cursor.              |
| ctrl+w           | Delete from cursor to start of the word.             |
| delete,ctrl+d    | Delete character to the right of cursor.             |
| ctrl+f           | Delete from cursor to end of the word.               |
| ctrl+x           | Delete the current line.                             |
| ctrl+u           | Delete from cursor to the start of the line.         |
| ctrl+k           | Delete from cursor to the end of the line.           |
| f6               | Select the current line.                             |
| f7               | Select all text in the document.                     |
"""
    _title = "Editor Keys Bindings"

    def compose(self) -> ComposeResult:
        """Compose the content of the modal dialog."""
        with Vertical(id="dialog"):
            yield MarkdownViewer(self.key_bindings, classes="spaced", show_table_of_contents=False)


class DefaultImportsInfo(ModalScreen[None]):
    BINDINGS = [
        Binding(
            "escape",
            "dismiss(None)",
            "Close",
        ),
    ]

    DEFAULT_CSS = """
    DefaultImportsInfo {
        align: center middle;
    }
"""

    _title = "Default Imported Modules"

    def __init__(
        self,
        imported_modules: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        self.imported_modules = imported_modules
        super().__init__(name, id, classes)

    def compose(self) -> ComposeResult:
        """Compose the content of the modal dialog."""
        syntax = Syntax(
            code=self.imported_modules,
            lexer="python",
            line_numbers=True,
            word_wrap=False,
            indent_guides=True,
            theme="dracula",
        )
        with VerticalScroll(id="dialog"):
            yield Label(syntax)


class InteractiveShellScreen(Screen):
    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        super().__init__(name, id, classes)
        self.input_tarea = ExtendedTextArea("", id="input", language="python", theme="vscode_dark")
        self.output_tarea = TextArea(
            "# Output",
            id="output",
            language="python",
            theme="vscode_dark",
            classes="text-area",
        )

    BINDINGS = [
        Binding(key="ctrl+r", action="run_code", description="Run the query"),
        Binding(key="ctrl+z", action="copy_command", description="Copy to Clipboard"),
        Binding(key="f1", action="editor_keys", description="Key Bindings"),
        Binding(key="f2", action="default_imports", description="Default imports"),
        Binding(key="ctrl+j", action="select_mode('commands')", description="Commands"),
        Binding(key="ctrl+underscore", action="toggle_comment", description="Toggle Comment", show=False),
    ]

    def compose(self) -> ComposeResult:
        self.input_tarea.focus()

        yield HorizontalScroll(
            self.input_tarea,
            self.output_tarea,
        )
        yield Label(f"Python: {platform.python_version()}  Django: {django.__version__}")
        yield Footer()

    def action_default_imports(self) -> None:
        self.app.push_screen(DefaultImportsInfo(import_str()))

    def action_run_code(self) -> None:
        # get Code from start till the position of the cursor
        self.input_tarea.selection = Selection(start=(0, 0), end=self.input_tarea.cursor_location)
        self.input_tarea.action_cursor_line_end()
        code = self.input_tarea.get_text_range(start=(0, 0), end=self.input_tarea.cursor_location)

        if len(code) > 0:
            # Because the cli - texualize is running on a loop - has an event loop
            # os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'rest.settings')
            os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
            django.setup()

            result = run_code(code)
            self.output_tarea.load_text(result["out"])

    def action_copy_command(self) -> None:
        if sys.platform == "win32":
            copy_command = ["clip"]
        elif sys.platform == "darwin":
            copy_command = ["pbcopy"]
        else:
            copy_command = ["xclip", "-selection", "clipboard"]

        try:
            # If nothing is selected copy all text in focused area
            if not self.input_tarea.selected_text and not self.output_tarea.selected_text:
                if self.input_tarea.text and self.input_tarea.has_focus:
                    text_to_copy = self.input_tarea.text
                    msg = "Input copied to clipboard."
                elif self.output_tarea.text and self.output_tarea.has_focus:
                    text_to_copy = self.output_tarea.text
                    msg = "Output copied to clipboard."
                else:
                    self.notify("Nothing to copy to clipboard.", severity="warning")
                    return

            # If both areas have selected text copy the selected text in focused area
            elif self.input_tarea.selected_text and self.output_tarea.selected_text:
                if self.input_tarea.has_focus:
                    text_to_copy = self.input_tarea.selected_text
                    msg = "Input selection copied to clipboard."
                else:
                    text_to_copy = self.output_tarea.selected_text
                    msg = "Output selection copied to clipboard."
            elif self.input_tarea.selected_text:
                text_to_copy = self.input_tarea.selected_text
                msg = "Input selection copied to clipboard."
            elif self.output_tarea.selected_text:
                text_to_copy = self.output_tarea.selected_text
                msg = "Output selection copied to clipboard."
            else:
                self.notify("Nothing to copy to clipboard.", severity="warning")
                return

            run(
                copy_command,
                input=text_to_copy,
                text=True,
                check=False,
            )
            self.notify(msg)
        except FileNotFoundError:
            self.notify(f"Could not copy to clipboard. `{copy_command[0]}` not found.", severity="error")

    def _get_selected_lines(self) -> Tuple[List[str], Location, Location]:
        [first, last] = sorted([self.input_tarea.selection.start, self.input_tarea.selection.end])
        lines = [self.input_tarea.document.get_line(i) for i in range(first[0], last[0] + 1)]
        return lines, first, last

    def action_toggle_comment(self) -> None:
        inline_comment_marker = "#"

        if inline_comment_marker:
            lines, first, last = self._get_selected_lines()
            stripped_lines = [line.lstrip() for line in lines]
            indents = [len(line) - len(line.lstrip()) for line in lines]
            # if lines are already commented, remove them
            if lines and all(not line or line.startswith(inline_comment_marker) for line in stripped_lines):
                offsets = [
                    0 if not line else (2 if line[len(inline_comment_marker)].isspace() else 1)
                    for line in stripped_lines
                ]
                for lno, indent, offset in zip(range(first[0], last[0] + 1), indents, offsets):
                    self.input_tarea.delete(
                        start=(lno, indent),
                        end=(lno, indent + offset),
                        maintain_selection_offset=True,
                    )
            # add comment tokens to all lines
            else:
                indent = min([indent for indent, line in zip(indents, stripped_lines) if line])
                for lno, stripped_line in enumerate(stripped_lines, start=first[0]):
                    if stripped_line:
                        self.input_tarea.insert(
                            f"{inline_comment_marker} ",
                            location=(lno, indent),
                            maintain_selection_offset=True,
                        )

    def action_editor_keys(self) -> None:
        self.app.push_screen(TextEditorBindingsInfo())
